// Scintilla source code edit control
// PlatTK.cxx - implementation of platform facilities on Linux using Tk
// Copyright 1998-2004 by Neil Hodgson <neilh@scintilla.org>
// The License.txt file describes the conditions under which this software may be distributed.
//
// Portions copied from
// tclWinLoad.c, tclLoadDl.c --
//
// Copyright (c) 1995-1997 Sun Microsystems, Inc.
//
// See the file "license.terms" for information on usage and redistribution of
// this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//
// Copyright (c) 2013 Mentor Graphics Corporation
// The license.terms file describes the conditions under which this software may be distributed.

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <math.h>

#include <vector>
#include <map>

#ifdef _WIN32
#include <windows.h>
#endif

#include <tk.h>

#ifdef SCI_LEXER
#ifndef _WIN32
/* we include tclInt.h to pick up the definition of TclpFindSymbol. This doesn't work
 * on Windows because the symbol isn't exported.
 */
extern "C" {
#include <tclInt.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
// tkUnixPort.h picks up the definition of TkPutImage
#include <tkUnixPort.h>
}
#endif

// dynamic loader also requires interpreter
extern Tcl_Interp *scintilla_interp;
#endif

#include "Platform.h"

#include "Scintilla.h"
#include "ScintillaTk.h"
#include "UniConversion.h"
#include "XPM.h"

#include "scintilla-ext.h"
using namespace TkSciExt;

#define USE_LOCK 0

#ifdef SCI_NAMESPACE
using namespace Scintilla;
#endif

enum encodingType { singleByte, UTF8, dbcs};

#ifdef SCI_NAMESPACE
namespace Scintilla {
#endif


#if USE_LOCK
static GMutex *fontMutex = NULL;

static void InitializeGLIBThreads() {
   if (!g_thread_supported()) {
      g_thread_init(NULL);
   }
}
#endif

static void FontMutexAllocate() {
#if USE_LOCK
   if (!fontMutex) {
      InitializeGLIBThreads();
      fontMutex = g_mutex_new();
   }
#endif
}

static void FontMutexFree() {
#if USE_LOCK
   if (fontMutex) {
      g_mutex_free(fontMutex);
      fontMutex = NULL;
   }
#endif
}

static void FontMutexLock() {
#if USE_LOCK
   g_mutex_lock(fontMutex);
#endif
}

static void FontMutexUnlock() {
#if USE_LOCK
   if (fontMutex) {
      g_mutex_unlock(fontMutex);
   }
#endif
}

#if !USE_SOLID_INDICATORS
static unsigned char *RGBAPixMap = NULL;
static size_t RGBAPixMapSize = 0;
#endif

// The following will encapsulate the platform-specific font
class FontHandle {
   int width[128];
   encodingType et;
   int refcount;    // refcount to reduce expensive calls to Tk_GetFont calls
public:
   friend class FontHandleMgr;
   Tk_Font tkfont;
   int characterSet;
   FontHandle(Tk_Font tkfont_) {
      et = singleByte;
      tkfont = tkfont_;
      characterSet = -1;
      ResetWidths(et);
      refcount = 0;
   }
   ~FontHandle() {
      Tk_FreeFont(tkfont);
   }
   void ResetWidths(encodingType et_) {
      et = et_;
      for (int i=0; i<=127; i++) {
         width[i] = 0;
      }
   }
   int CharWidth(unsigned char ch, encodingType et_) {
      int w = 0;
      FontMutexLock();
      if ((ch <= 127) && (et == et_)) {
         w = width[ch];
      }
      FontMutexUnlock();
      return w;
   }
   void SetCharWidth(unsigned char ch, int w, encodingType et_) {
      if (ch <= 127) {
         FontMutexLock();
         if (et != et_) {
            ResetWidths(et_);
         }
         width[ch] = w;
         FontMutexUnlock();
      }
   }
};

// reference counting manager (smart pointer for FontHandle Objects
class FontHandleMgr {
private:
    FontHandle *handle_p;
public:
    FontHandle* GetHandle()  { return handle_p; }
    FontHandle* operator->() { return handle_p; }
    FontHandle& operator*() { return *handle_p; }
    FontHandleMgr(FontHandle*p) { handle_p = p; ++p->refcount; }
    ~FontHandleMgr() {
        if (--handle_p->refcount == 0) {
            delete handle_p;
        }
    }
    FontHandleMgr(FontHandleMgr const &other) { handle_p = other.handle_p; ++handle_p->refcount; }
    FontHandleMgr & operator= (FontHandleMgr const &other) {
        // DO NOT CHANGE THE ORDER OF THESE STATEMENTS!
        // (This order properly handles self-assignment)
        // (This order also properly handles recursion, e.g., if a FontHandle contains FontHandleMgr)
        FontHandle* const old = handle_p;
        handle_p = other.handle_p;
        ++handle_p->refcount;
        if (--old->refcount == 0) {
            delete old;
        }
        return *this;
    }
};

// X has a 16 bit coordinate space, so stop drawing here to avoid wrapping
static const int maxCoordinate = 32000;

static FontHandle *PFont(Font &f) {
	//return reinterpret_cast<FontHandle *>(f.GetID());
	FontHandle *h = (reinterpret_cast<FontHandleMgr *>(f.GetID()))->GetHandle();
    return h;
}


///////////////////////////////////////////////////////////
// CLASS: SurfaceImpl
///////////////////////////////////////////////////////////
class SurfaceImpl : public Surface {
   encodingType et;
   Tk_Window tkwin;
   Tcl_Interp *interp;
   ScintillaObject *sc_obj;
   Display *display;
   GC gc;
   GC copyGC;
   Pixmap pixmap;

   int x;
   int y;
   bool inited;
   bool mainWindow; // TRUE if this is the main(client) window
   int mainWidth;
   int mainHeight;

   int characterSet;
   void SetConverter(int characterSet_);
   void FreeGC();
public:
   SurfaceImpl();
   virtual ~SurfaceImpl();

    void Init(WindowID wid);
    void Init(SurfaceID sid, WindowID wid);
    void InitPixMap(int width, int height, Surface *surface_, WindowID wid);

    void Release();
    bool Initialised();
    void PenColour(ColourDesired fore);
    int LogPixelsY();
    int DeviceHeightFont(int points);
    void MoveTo(int x_, int y_);
    void LineTo(int x_, int y_);
    void Polygon(Point *pts, int npts, ColourDesired fore, ColourDesired back);
    void RectangleDraw(PRectangle rc, ColourDesired fore, ColourDesired back);
    void FillRectangle(PRectangle rc, ColourDesired back);
    void FillRectangle(PRectangle rc, Surface &surfacePattern);
    void RoundedRectangle(PRectangle rc, ColourDesired fore, ColourDesired back);
    void AlphaRectangle(PRectangle rc, int cornerSize, ColourDesired fill, int alphaFill,
                        ColourDesired outline, int alphaOutline, int flags);
    void DrawRGBAImage(PRectangle rc, int width, int height, const unsigned char *pixelsImage);
    void Ellipse(PRectangle rc, ColourDesired fore, ColourDesired back);
    void Copy(PRectangle rc, Point from, Surface &surfaceSource);

    void DrawTextBase(PRectangle rc, Font &font_, XYPOSITION ybase, const char *s, int len, ColourDesired fore);
    void DrawTextNoClip(PRectangle rc, Font &font_, XYPOSITION ybase, const char *s, int len, ColourDesired fore, ColourDesired back);
    void DrawTextClipped(PRectangle rc, Font &font_, XYPOSITION ybase, const char *s, int len, ColourDesired fore, ColourDesired back);
    void DrawTextTransparent(PRectangle rc, Font &font_, XYPOSITION ybase, const char *s, int len, ColourDesired fore);
    void MeasureWidths(Font &font_, const char *s, int len, XYPOSITION *positions);
    XYPOSITION WidthText(Font &font_, const char *s, int len);
    XYPOSITION WidthChar(Font &font_, char ch);
    XYPOSITION Ascent(Font &font_);
    XYPOSITION Descent(Font &font_);
    XYPOSITION InternalLeading(Font &font_);
    XYPOSITION ExternalLeading(Font &font_);
    XYPOSITION Height(Font &font_);
    XYPOSITION AverageCharWidth(Font &font_);

    void SetClip(PRectangle rc);
    void FlushCachedState();

    void SetUnicodeMode(bool unicodeMode_);
    void SetDBCSMode(int codePage);
private:
    void ImgBlendComplexAlpha(
        XImage *bgImg,                    /* Background image to draw on. */
        const unsigned char *pixelsImage, /* Image instance to draw. */
        int xOffset, int yOffset,         /* X & Y offset into image instance to
                                           * draw. */
        int width, int height);           /* Width & height of image to draw. */
    XImage *GetXImage(PRectangle rc, Tk_ErrorHandler handler);
};
#ifdef SCI_NAMESPACE
}
#endif

const char *CharacterSetID(int characterSet) {
   switch (characterSet) {
   case SC_CHARSET_ANSI:
      return "";
   case SC_CHARSET_DEFAULT:
      return "ISO-8859-1";
   case SC_CHARSET_BALTIC:
      return "ISO-8859-13";
   case SC_CHARSET_CHINESEBIG5:
      return "BIG-5";
   case SC_CHARSET_EASTEUROPE:
      return "ISO-8859-2";
   case SC_CHARSET_GB2312:
      return "CP936";
   case SC_CHARSET_GREEK:
      return "ISO-8859-7";
   case SC_CHARSET_HANGUL:
      return "CP949";
   case SC_CHARSET_MAC:
      return "MACINTOSH";
   case SC_CHARSET_OEM:
      return "ASCII";
   case SC_CHARSET_RUSSIAN:
      return "KOI8-R";
   case SC_CHARSET_CYRILLIC:
      return "CP1251";
   case SC_CHARSET_SHIFTJIS:
      return "SHIFT-JIS";
   case SC_CHARSET_SYMBOL:
      return "";
   case SC_CHARSET_TURKISH:
      return "ISO-8859-9";
   case SC_CHARSET_JOHAB:
      return "CP1361";
   case SC_CHARSET_HEBREW:
      return "ISO-8859-8";
   case SC_CHARSET_ARABIC:
      return "ISO-8859-6";
   case SC_CHARSET_VIETNAMESE:
      return "";
   case SC_CHARSET_THAI:
      return "ISO-8859-11";
   case SC_CHARSET_8859_15:
      return "ISO-8859-15";
   default:
      return "";
   }
}

void SurfaceImpl::SetConverter(int characterSet_) {
//    if (characterSet != characterSet_) {
//       characterSet = characterSet_;
//       conv.Open("UTF-8", CharacterSetID(characterSet), false);
//    }
}

SurfaceImpl::SurfaceImpl() : et(singleByte),
                             sc_obj(0),
                             display(0),
                             gc(0), copyGC(0),
                             pixmap(0),
                             x(0), y(0), 
                             inited(false),
                             mainWindow(false),mainWidth(0),mainHeight(0),
                             characterSet(-1) {
}

SurfaceImpl::~SurfaceImpl() {
//printf("Surface DTOR(main=%d) -- THIS=%p\n",mainWindow,(void *)this);
   FreeGC();
   if (pixmap && display) {
      Tk_FreePixmap(display, pixmap);
   }
}

/*
 *----------------------------------------------------------------------
 * Release --
 *
 *    Handles releases resources used by this surface
 *
 * Results:
 *    None
 *
 * Side affects:
 *    Resources releases and flags are reset
 *
 *----------------------------------------------------------------------
 */
void SurfaceImpl::Release() {
#if 0
   et = singleByte;
//printf("---RELEASE(main=%d) %p\n",mainWindow,(void *)this);
//_print_trace(1);
   FreeGC();

   if (pixmap && display) {
      Tk_FreePixmap(display, pixmap);
      pixmap = 0;
   }
#endif
   characterSet = -1;
   x = 0;
   y = 0;
   inited = false;
}
void SurfaceImpl::FreeGC() {
   if (gc) {
//printf("FreeGC (%p) >>> %d\n",(void *)this,gc);
      Tk_FreeGC(display, gc);
      gc = 0;
   }
}

bool SurfaceImpl::Initialised() {
//printf("SurfaceImpl::Initialised THIS=%p  inited=%d  pixmap=%d\n",(void *)this,inited,(int)pixmap);
   return (inited && (pixmap != 0));
}
static Tk_Window _curr_tkwin = 0;

/*
 *----------------------------------------------------------------------
 * Init --
 *
 *    Called by Editor to get the drawing pixmap for one of the internal
 *    surfaces initialized.
 *
 * Results:
 *    None
 *
 * Side affects:
 *    Creates drawing pixmap with requested dimensions
 *
 *----------------------------------------------------------------------
 */
void SurfaceImpl::Init(WindowID wid) {
   PLATFORM_ASSERT(wid);
   sc_obj = reinterpret_cast<ScintillaObject *>(wid);
   mainWindow = sc_obj->main;
//printf("SurfaceImpl::Init(main=%d) THIS=%p  WID=%p\n",mainWindow,(void *)this,(void *)wid);
   if (!mainWindow) {
      Release();
   }
   tkwin = sc_obj->tkwin;
//TODO: figure how to avoid needed the next line!!!!
   _curr_tkwin = tkwin;

   interp  = Tk_Interp(tkwin);
   display = Tk_Display(tkwin);

   if (mainWindow && Tk_WindowId(tkwin)) {
      int new_w = Tk_Width(tkwin);
      int new_h = Tk_Height(tkwin);
      if ((new_w != mainWidth) || (new_h != mainHeight)) {
         if (pixmap)
            Tk_FreePixmap(display, pixmap);
         pixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin),
            new_w, new_h, Tk_Depth(tkwin));
         mainWidth  = new_w;
         mainHeight = new_h;
//printf("SurfaceImpl::Init THIS=%p  WID=%p  PIXMAP=%d  WIDTH=%d HEIGHT=%d\n",(void *)this,(void *)wid,(int)pixmap,Tk_Width(tkwin),Tk_Height(tkwin));
      }
   }
   inited = true;
}

void SurfaceImpl::Init(SurfaceID sid, WindowID wid) {
   // it's not clear why we need to support this API, but since we're
   // required to implement it, just call the simple version
   Init(wid);
}

/*
 *----------------------------------------------------------------------
 * InitPixMap --
 *
 *    Called by Editor to get the drawing pixmap for one of the internal
 *    surfaces initialized. The surface which is passed in corresponds
 *    to the main(client) window.
 *
 * Results:
 *    None
 *
 * Side affects:
 *    Creates drawing pixmap with requested dimensions
 *
 *----------------------------------------------------------------------
 */
void SurfaceImpl::InitPixMap(int width, int height, Surface *surface_, WindowID wid) {
   PLATFORM_ASSERT(surface_);
   sc_obj = reinterpret_cast<ScintillaObject *>(wid);
   Release();
   SurfaceImpl *surfImpl = static_cast<SurfaceImpl *>(surface_);
//printf("INIT PIXMAP(main=%d, wid=%p): THIS=%p SRC=%p width=%d height=%d\n",sc_obj->main,(void *)wid,(void *)this,(void *)surfImpl,width,height);

   if (height > 0 && width > 0) {
      if (pixmap)
         Tk_FreePixmap(Tk_Display(tkwin), pixmap);
      pixmap = Tk_GetPixmap( surfImpl->display, Tk_WindowId(surfImpl->tkwin),
         width, height, Tk_Depth(surfImpl->tkwin) );
//printf("---InitPixMap: THIS=%p SRC=%p width=%d height=%d  PIX=%d\n",(void *)this,(void *)surfImpl,width,height,(int)pixmap);
   }
#if 1
   tkwin   = sc_obj->tkwin;
   interp  = Tk_Interp(tkwin);
   display = Tk_Display(tkwin);
   inited  = true;
#else
   if ( ! Initialised() ) {
      Init(wid);
   }
#endif
}

/*
 *----------------------------------------------------------------------
 * PenColour --
 *
 *    Called prior to drawing operation to set a specified color
 *
 * Results:
 *    None
 *
 * Side affects:
 *    A GC is obtained (Tk is allowed to reuse) with the specified
 *    color defined for the foreground.
 *
 *----------------------------------------------------------------------
 */
void SurfaceImpl::PenColour(ColourDesired fore) {
   FreeGC();

   XGCValues gcvalues;
   XColor *xc;
   static char rgb[7];
   sprintf(rgb, "#%02x%02x%02x", fore.GetRed(), fore.GetGreen(), fore.GetBlue());
   xc = Tk_GetColor(interp, tkwin, rgb);
   if (xc) {
      gcvalues.foreground = xc->pixel;
      gc = Tk_GetGC(tkwin, GCForeground, &gcvalues);
//printf("PenColour ####### SET GC = %ld for THIS=%p RGB=%d,%d,%d\n",(long)gc,(void *)this,fore.GetRed(),fore.GetGreen(),fore.GetBlue());
      Tk_FreeColor(xc);
   }
}

int SurfaceImpl::LogPixelsY() {
   return 72;
}

int SurfaceImpl::DeviceHeightFont(int points) {
   int logPix = LogPixelsY();
   return (points * logPix + logPix / 2) / 72;
}

void SurfaceImpl::MoveTo(int x_, int y_) {
//printf("MOVETO: %d,%d\n",x_,y_);
   x = x_;
   y = y_;
}

void SurfaceImpl::LineTo(int x_, int y_) {
   if (gc) {
//printf("LINETO: %d,%d -> %d,%d\n",x_,y_,x,y);
      XDrawLine( display, pixmap, gc, x_, y_, x, y );
   }
   x = x_;
   y = y_;
}

void SurfaceImpl::Polygon(Point *pts, int npts, ColourDesired fore, ColourDesired back) {
#if 0
printf("SurfaceImpl::Polygon  BG=%d,%d,%d  FG=%d,%d,%d\n",
   back.GetRed(),back.GetGreen(),back.GetBlue(), fore.GetRed(),fore.GetGreen(),fore.GetBlue());
#endif
   XPoint points[20];
   if (npts < static_cast<int>((sizeof(points) / sizeof(points[0])))) {
      int i;
      int _x = 0;
      int _y = 0;
      for (i = 0;i < npts;i++) {
         if (i==0) {
            _x = static_cast<int>(pts[i].x);
            _y = static_cast<int>(pts[i].y);
         }
         points[i].x = static_cast<short>(pts[i].x);
         points[i].y = static_cast<short>(pts[i].y);
      }
      // add the closing coordinate
      points[i].x = _x;
      points[i].y = _y;
      npts++;

      PenColour(back);
      XFillPolygon(display, pixmap, gc, points, npts, Convex, CoordModeOrigin);

      PenColour(fore);
      XDrawLines(display, pixmap, gc, points, npts, CoordModeOrigin);
   }
}

void SurfaceImpl::RectangleDraw(PRectangle rc, ColourDesired fore, ColourDesired back) {
#if 0
printf("##################################\n");
printf("# SurfaceImpl::RectangleDraw\n");
printf("#  FORE=%d,%d,%d",  fore.GetRed(),fore.GetGreen(),fore.GetBlue());
printf("   BACK=%d,%d,%d\n",back.GetRed(),back.GetGreen(),back.GetBlue());
printf("##################################\n");
#endif
   int x = static_cast<int>(rc.left);
   int y = static_cast<int>(rc.top);
   int w = static_cast<int>(rc.right - rc.left);
   int h = static_cast<int>(rc.bottom - rc.top);
   PenColour(back);
   XDrawRectangle(display, pixmap, gc, x + 1, y + 1, w - 2, h - 2);

   PenColour(fore);

   // The subtraction of 1 off the width and height here shouldn't be needed but
   // otherwise a different rectangle is drawn than would be done if the fill
   // parameter == 1

   XDrawRectangle(display, pixmap, gc, x, y, w - 1, h -1);
}

void SurfaceImpl::FillRectangle(PRectangle rc, ColourDesired back) {
#if 0
if ((rc.right-rc.left)>1 && (rc.bottom-rc.top)>1)
   printf("\nFILL RECT1 RGB=%d,%d,%d THIS=%p|PIX=%d|GC=%ld|+%d+%dx%dx%d\n",
      back.GetRed(),back.GetGreen(),back.GetBlue(),
      (void *)this,(int)pixmap,(long)gc,
      rc.left,rc.top,(rc.right-rc.left),(rc.bottom-rc.top));
#endif
   PenColour(back);
   if (gc && (rc.left < maxCoordinate)) { // Protect against out of range
//////////////////////
//-------CHANGE TO : Tk_Fill3DRectangle --- use Tk_3DBorder
      bool to_screen = false;

      if (to_screen) {
      XFillRectangle(display, Tk_WindowId(tkwin), gc,
         static_cast<int>(rc.left), static_cast<int>(rc.top),
         static_cast<int>(rc.right - rc.left), static_cast<int>(rc.bottom - rc.top));
      Tcl_DoOneEvent(2); //DRAW NOW!!!!

      } else {
      XFillRectangle(display, pixmap, gc,
         static_cast<int>(rc.left), static_cast<int>(rc.top),
         static_cast<int>(rc.right - rc.left), static_cast<int>(rc.bottom - rc.top));
      }
   }
}

void SurfaceImpl::FillRectangle(PRectangle rc, Surface &surfacePattern) {
//printf("SurfaceImpl::FillRectangle 2  THIS=%p\n",(void *)this);
   SurfaceImpl &surfOther = static_cast<SurfaceImpl &>(surfacePattern);
   if (Initialised() && surfOther.Initialised()) {
      // Tile pattern over rectangle
      // Currently assumes 8x8 pattern
      int widthPat = 8;
      int heightPat = 8;
      for (int xTile = static_cast<int>(rc.left); xTile < rc.right; xTile += widthPat) {
         int widthx = (xTile + widthPat > rc.right) ? static_cast<int>(rc.right - xTile) : widthPat;
         for (int yTile = static_cast<int>(rc.top); yTile < rc.bottom; yTile += heightPat) {
            int heighty = (yTile + heightPat > rc.bottom) ? static_cast<int>(rc.bottom - yTile) : heightPat;

            if (!gc) {
               XGCValues gcValues; // just use defaults
               gc = Tk_GetGC(tkwin, GCGraphicsExposures, &gcValues);
            }
            XCopyArea(display, surfOther.pixmap, pixmap, gc,
               0, 0,        /* src X/Y */
               widthx, heighty,
               xTile, yTile /* dest X/Y */
            );
         }
      }
   } else {
      // Something is wrong so try to show anyway
      // Shows up black because colour not allocated
      FillRectangle(rc, ColourDesired(0));
   }
}

void SurfaceImpl::RoundedRectangle(PRectangle rc, ColourDesired fore, ColourDesired back) {
   if (((rc.right - rc.left) > 4) && ((rc.bottom - rc.top) > 4)) {
       // Approximate a round rect with some cut off corners
       Point pts[] = {
           Point(rc.left + 2,  rc.top),
           Point(rc.right - 2, rc.top),
           Point(rc.right,     rc.top + 2),
           Point(rc.right,     rc.bottom - 3),
           Point(rc.right - 3, rc.bottom),
           Point(rc.left + 3,  rc.bottom),
           Point(rc.left,      rc.bottom - 3),
           Point(rc.left,      rc.top + 2),
       };
       Polygon(pts, sizeof(pts) / sizeof(pts[0]), fore, back);
   } else {
       RectangleDraw(rc, fore, back);
   }
}

inline int CountBits(int mask)  /* Value to count the 1 bits in. */
{
    int n;

    for (n=0 ; mask!=0 ; mask&=mask-1) {
	n++;
    }
    return n;
}

XImage *SurfaceImpl::GetXImage(PRectangle rc, Tk_ErrorHandler handler) {
    int nreturned;
    XImage *bgImg = NULL;
    XVisualInfo * visInfo = XGetVisualInfo(display, 0, NULL, &nreturned);
    if ((visInfo->depth >= 15) && (visInfo->c_class == DirectColor || visInfo->c_class == TrueColor)) {
        int drawableX = static_cast<int>(rc.left);
        int drawableY = static_cast<int>(rc.top);
        int drawableWidth = static_cast<int>(rc.right - rc.left);
        int drawableHeight = static_cast<int>(rc.bottom - rc.top);
	    /*
	     * Pull the current background from the display to blend with
	     */

	    bgImg = XGetImage(display, pixmap /*Tk_WindowId(tkwin)*/, drawableX, drawableY,
		    (unsigned int)drawableWidth, (unsigned int)drawableHeight, AllPlanes, ZPixmap);
    }
    return bgImg;
}

void SurfaceImpl::AlphaRectangle(PRectangle rc, int cornerSize, ColourDesired fill, int alphaFill,
      ColourDesired outline, int alphaOutline, int flags) {
#if USE_SOLID_INDICATORS
    if (cornerSize)
        RoundedRectangle(rc, fill, fill);
    else
        FillRectangle(rc, fill);
#else
int drawableX = static_cast<int>(rc.left);
int drawableY = static_cast<int>(rc.top);
int width = static_cast<int>(rc.right - rc.left);
int height = static_cast<int>(rc.bottom - rc.top);

    /*
     * Create an error handler to suppress the case where the input was
     * not properly constrained, which can cause an X error. [Bug 979239]
     */

    Tk_ErrorHandler handler = Tk_CreateErrorHandler(display, -1, -1, -1, NULL, (ClientData) NULL);

    XImage *bgImg = GetXImage(rc, handler);
    if (bgImg) {
        int x, y;
        unsigned char *p;
        size_t desiredSize = 4*width*height;

        if (RGBAPixMapSize < desiredSize) {
            RGBAPixMap = (unsigned char*)realloc(RGBAPixMap, desiredSize);
            RGBAPixMapSize = desiredSize;
        }
        // build RGBA rectangle
        p = RGBAPixMap;
        for (y = 0; y < height; ++y) {
            for (x = 0; x < width; ++x) {
                if ((x==0) || (x==width-1) || (y == 0) || (y == height-1)) {
                    // boarder pixel
                    p[0] = outline.GetRed();
                    p[1] = outline.GetGreen();
                    p[2] = outline.GetBlue();
                    p[3] = alphaOutline;
                } else {
                    p[0] = fill.GetRed();
                    p[1] = fill.GetGreen();
                    p[2] = fill.GetBlue();
                    p[3] = alphaFill;
                }
                p += 4;
            }
        }
        if (cornerSize && (width > 4) && (height > 4)) {
            // round the corners of the rectangle by adjusting the alpha values
            p = RGBAPixMap;
            p[3] = 0; /* make top left two pixels transparent */
            p += 4;
            p[3] = 0;
            p += (4 * (width - 3)); /* now the top right two pixels */
            p[3] = 0;
            p += 4; 
            p[3] = 0;
            p += 4; /* now the first pixel on the left, one line down */
            p[3] = 0;
            p += 4; /* now the next pixel should be an outline pixel */
            p[0] = outline.GetRed();
            p[1] = outline.GetGreen();
            p[2] = outline.GetBlue();
            p[3] = alphaOutline;
            p += (4 * (width - 3)); /* then the 2nd from the right pixel, one line down */
            p[0] = outline.GetRed();
            p[1] = outline.GetGreen();
            p[2] = outline.GetBlue();
            p[3] = alphaOutline;
            p += 4; 
            p[3] = 0;
            p += 4 + (4 * (height - 4) * width); /* then the left pixel, two from the bottom */
            p[3] = 0;
            p += 4; 
            p[0] = outline.GetRed();
            p[1] = outline.GetGreen();
            p[2] = outline.GetBlue();
            p[3] = alphaOutline;
            p += (4 * (width - 3));
            p[0] = outline.GetRed();
            p[1] = outline.GetGreen();
            p[2] = outline.GetBlue();
            p[3] = alphaOutline;
            p += 4; 
            p[3] = 0;
            p += 4;
            p[3] = 0;
            p += 4;
            p[3] = 0;
            p += (4 * (width - 3));
            p[3] = 0;
            p += 4;
            p[3] = 0;
        }
        ImgBlendComplexAlpha(bgImg, RGBAPixMap, 0, 0, width,  height);
        TkPutImage(NULL, 0, display, pixmap/*Tk_WindowId(tkwin)*/, gc, bgImg, 0, 0, drawableX, drawableY, 
              (unsigned int)width, (unsigned int)height);
        XDestroyImage(bgImg);
    }
    Tk_DeleteErrorHandler(handler);
#endif
}

#ifndef __WIN32__
#define GetRValue(rgb)	(UCHAR(((rgb) & red_mask) >> red_shift))
#define GetGValue(rgb)	(UCHAR(((rgb) & green_mask) >> green_shift))
#define GetBValue(rgb)	(UCHAR(((rgb) & blue_mask) >> blue_shift))
#define RGB(r, g, b)	((unsigned)( \
	(UCHAR(r) << red_shift)   | \
	(UCHAR(g) << green_shift) | \
	(UCHAR(b) << blue_shift)  ))
#define RGB15(r, g, b)	((unsigned)( \
	(((r) * red_mask / 255)   & red_mask)   | \
	(((g) * green_mask / 255) & green_mask) | \
	(((b) * blue_mask / 255)  & blue_mask)  ))
#endif /* !__WIN32__ */

void SurfaceImpl::ImgBlendComplexAlpha(
    XImage *bgImg,                    /* Background image to draw on. */
    const unsigned char *pixelsImage, /* Image instance to draw. NULL means color fill */
    int xOffset, int yOffset,         /* X & Y offset into image instance to
                                       * draw. */
    int width, int height)            /* Width & height of image to draw. */
{
    int x, y, line;
    unsigned long pixel;
    unsigned char r, g, b, alpha, unalpha;
    const unsigned char *masterPtr = NULL;
    const unsigned char *alphaAr = pixelsImage;

    /*
     * This blending is an integer version of the Source-Over compositing rule
     * (see Porter&Duff, "Compositing Digital Images", proceedings of SIGGRAPH
     * 1984) that has been hard-coded (for speed) to work with targetting a
     * solid surface.
     *
     * The 'unalpha' field must be 255-alpha; it is separated out to encourage
     * more efficient compilation.
     */

#define ALPHA_BLEND(bgPix, imgPix, alpha, unalpha) \
	((bgPix * unalpha + imgPix * alpha) / 255)

    /*
     * We have to get the mask and shift info from the visual on non-Win32 so
     * that the macros Get*Value(), RGB() and RGB15() work correctly. This
     * might be cached for better performance.
     */

#ifndef __WIN32__
    unsigned long red_mask, green_mask, blue_mask;
    unsigned long red_shift, green_shift, blue_shift;
    int depth;
    Colormap colormap;
    Visual *visual = Tk_GetVisual(interp, tkwin, "default", &depth, &colormap);

    red_mask = visual->red_mask;
    green_mask = visual->green_mask;
    blue_mask = visual->blue_mask;
    red_shift = 0;
    green_shift = 0;
    blue_shift = 0;
    while ((0x0001 & (red_mask >> red_shift)) == 0) {
	red_shift++;
    }
    while ((0x0001 & (green_mask >> green_shift)) == 0) {
	green_shift++;
    }
    while ((0x0001 & (blue_mask >> blue_shift)) == 0) {
	blue_shift++;
    }
#endif /* !__WIN32__ */

    /*
     * Only UNIX requires the special case for <24bpp. It varies with 3 extra
     * shifts and uses RGB15. The 24+bpp version could also then be further
     * optimized.
     */

#if !(defined(__WIN32__) || defined(MAC_OSX_TK))
    if (bgImg->depth < 24) {
	unsigned char red_mlen, green_mlen, blue_mlen;

	red_mlen = 8 - CountBits(red_mask >> red_shift);
	green_mlen = 8 - CountBits(green_mask >> green_shift);
	blue_mlen = 8 - CountBits(blue_mask >> blue_shift);
	for (y = 0; y < height; y++) {
	    line = (y + yOffset) * width;
	    for (x = 0; x < width; x++) {

        /* if we're merging images, get the RGBA value for the current pixel */
        masterPtr = alphaAr + ((line + x + xOffset) * 4);
        alpha = masterPtr[3];

		/*
		 * Ignore pixels that are fully transparent
		 */

		if (alpha) {
		    /*
		     * We could perhaps be more efficient than XGetPixel for
		     * 24 and 32 bit displays, but this seems "fast enough".
		     */

            r = masterPtr[0];
            g = masterPtr[1];
            b = masterPtr[2];

            if (alpha != 255) {
                /*
                 * Only blend pixels that have some transparency
                 */
                unsigned char ra, ga, ba;

                pixel = XGetPixel(bgImg, x, y);
                ra = GetRValue(pixel) << red_mlen;
                ga = GetGValue(pixel) << green_mlen;
                ba = GetBValue(pixel) << blue_mlen;
                unalpha = 255 - alpha;	/* Calculate once. */
                r = ALPHA_BLEND(ra, r, alpha, unalpha);
                g = ALPHA_BLEND(ga, g, alpha, unalpha);
                b = ALPHA_BLEND(ba, b, alpha, unalpha);
            }
		    XPutPixel(bgImg, x, y, RGB15(r, g, b));
		}
	    }
	}
	return;
    }
#endif /* !__WIN32__ && !MAC_OSX_TK */
    for (y = 0; y < height; y++) {
	line = (y + yOffset) * width;
	for (x = 0; x < width; x++) {
        /* if we're merging images, get the RGBA value for the current pixel */
        masterPtr = alphaAr + ((line + x + xOffset) * 4);
        alpha = masterPtr[3];

	    /*
	     * Ignore pixels that are fully transparent
	     */

	    if (alpha) {
		/*
		 * We could perhaps be more efficient than XGetPixel for 24
		 * and 32 bit displays, but this seems "fast enough".
		 */

        r = masterPtr[0];
        g = masterPtr[1];
        b = masterPtr[2];
		if (alpha != 255) {
		    /*
		     * Only blend pixels that have some transparency
		     */
		    unsigned char ra, ga, ba;

		    pixel = XGetPixel(bgImg, x, y);
		    ra = GetRValue(pixel);
		    ga = GetGValue(pixel);
		    ba = GetBValue(pixel);
		    unalpha = 255 - alpha;	/* Calculate once. */
		    r = ALPHA_BLEND(ra, r, alpha, unalpha);
		    g = ALPHA_BLEND(ga, g, alpha, unalpha);
		    b = ALPHA_BLEND(ba, b, alpha, unalpha);
		}
		XPutPixel(bgImg, x, y, RGB(r, g, b));
	    }
	}
    }
#undef ALPHA_BLEND
}

void SurfaceImpl::DrawRGBAImage(PRectangle rc, int width, int height, const unsigned char *pixelsImage) {
    int drawableX = static_cast<int>(rc.left);
    int drawableY = static_cast<int>(rc.top);

    /*
     * If there's no pixmap, it means that an error occurred while creating
     * the image instance so it can't be displayed.
     */

    if (pixelsImage == NULL) {
        fprintf(stderr, "pixelsImage = %p\n", pixelsImage);
        return;
    }

	/*
	 * Create an error handler to suppress the case where the input was
	 * not properly constrained, which can cause an X error. [Bug 979239]
	 */

	Tk_ErrorHandler handler = Tk_CreateErrorHandler(display, -1, -1, -1, NULL,
		(ClientData) NULL);

	/*
	 * Pull the current background from the display to blend with
	 */

	XImage *bgImg = GetXImage(rc, handler);
	if (bgImg != NULL) {
	    ImgBlendComplexAlpha(bgImg, pixelsImage, 0, 0, width, height);

	    /*
	     * Color info is unimportant as we only do this operation for depth >=
	     * 15.
	     */
        TkPutImage(NULL, 0, display, pixmap/*Tk_WindowId(tkwin)*/, gc, 
                  bgImg, 0, 0, drawableX, drawableY, 
                  (unsigned int)width, (unsigned int)height);
	    XDestroyImage(bgImg);
    }
	Tk_DeleteErrorHandler(handler);
}

void SurfaceImpl::Ellipse(PRectangle rc, ColourDesired fore, ColourDesired back) {

   static int full_circle = 64 * 360;
   PenColour(back);
   XFillArc(display, pixmap, gc,
      static_cast<int>(rc.left + 1), static_cast<int>(rc.top + 1),
      static_cast<int>(rc.right - rc.left - 2), static_cast<int>(rc.bottom - rc.top - 2),
      0, full_circle);

   // The subtraction of 1 here is similar to the case for RectangleDraw
   PenColour(fore);
   XDrawArc(display, pixmap, gc,
      static_cast<int>(rc.left), static_cast<int>(rc.top),
      static_cast<int>(rc.right - rc.left - 1), static_cast<int>(rc.bottom - rc.top - 1),
      0, full_circle);
}

/****************************************************************
// Func: Copy
// Desc: Called from Editor code to copy a region from a source
//       surface to the screen as a result of a paint action
*****************************************************************/
//int cnt = 0;
void SurfaceImpl::Copy(PRectangle rc, Point from, Surface &surfaceSource) {
   bool to_screen;
   GC _gc;

   SurfaceImpl &surfOther = static_cast<SurfaceImpl &>(surfaceSource);
   //printf("\nSurfaceImpl::Copy THIS=%p<%d> OTHER=%p<%d>\n",(void *)this,(int)pixmap,(void *)&surfOther,(int)surfOther.pixmap);
   if (!Initialised() || !surfOther.Initialised()) {
      return;
   }

   to_screen = (surfOther.pixmap == pixmap);

   if (to_screen) {
      if (!copyGC) {
         XGCValues gcValues;
         gcValues.graphics_exposures = True;
         copyGC = Tk_GetGC(tkwin, GCGraphicsExposures, &gcValues);
      }
      _gc = copyGC;
   } else {
      _gc = surfOther.gc;
   }

//   cnt++;
   if (to_screen)  {
	  TkSciExt::SetPixmap((ClientData)sc_obj->client, pixmap);
//printf(">>>>>XCopyArea %d FROM %02d %02d SCREEN +%02d+%02dx%dx%d \n",cnt, from.x, from.y,rc.left,rc.top,(rc.right-rc.left),(rc.bottom-rc.top));
      XCopyArea(display,surfOther.pixmap,Tk_WindowId(tkwin),_gc,static_cast<int>(rc.left), static_cast<int>(rc.top),static_cast<int>(rc.right - rc.left), static_cast<int>(rc.bottom - rc.top),static_cast<int>(rc.left), static_cast<int>(rc.top));
   } else {
//printf(">>>>>XCopyArea %d FROM %02d %02d BUFFER +%02d+%02dx%dx%d \n",cnt, from.x, from.y, rc.left,rc.top,(rc.right-rc.left),(rc.bottom-rc.top));
      XCopyArea(display,surfOther.pixmap, pixmap, _gc,static_cast<int>(from.x), static_cast<int>(from.y),static_cast<int>(rc.right - rc.left), static_cast<int>(rc.bottom - rc.top), static_cast<int>(rc.left), static_cast<int>(rc.top)); 
   }
}

char *UTF8FromLatin1(const char *s, int &len) {
   char *utfForm = new char[len*2+1];
   size_t lenU = 0;
   for (int i=0;i<len;i++) {
      unsigned int uch = static_cast<unsigned char>(s[i]);
      if (uch < 0x80) {
         utfForm[lenU++] = uch;
      } else {
         utfForm[lenU++] = static_cast<char>(0xC0 | (uch >> 6));
         utfForm[lenU++] = static_cast<char>(0x80 | (uch & 0x3f));
      }
   }
   utfForm[lenU] = '\0';
   len = lenU;
   return utfForm;
}

void SurfaceImpl::DrawTextBase(
   PRectangle rc, Scintilla::Font &font_, XYPOSITION ybase, const char *s, int len, ColourDesired fore
) {
//printf("DRAW TEXT(%d): @%d,%d THIS=%p pixmap=%d GC=%ld  STR=%s\n",len,rc.left,ybase,(void *)this,(int)pixmap,(long)gc,s);

   PenColour(fore);
   if (pixmap && gc && PFont(font_)) {
      Tk_Font tkfont = PFont(font_)->tkfont;
#ifdef _WIN32
      gc->font = (unsigned long)tkfont;
#endif
      Tk_DrawChars(display, pixmap, gc, tkfont, s, len, static_cast<int>(rc.left), static_cast<int>(ybase));
   } else {
//printf("DRAW TEXT ############### SKIPPED  pixmap=%d  gc=%ld\n",(int)pixmap,(long)gc);
   }
}

void SurfaceImpl::DrawTextNoClip(PRectangle rc, Scintilla::Font &font_, XYPOSITION ybase, const char *s, int len,
                                 ColourDesired fore, ColourDesired back) {
   FillRectangle(rc, back);
   DrawTextBase(rc, font_, ybase, s, len, fore);
}

// On GTK+, exactly same as DrawTextNoClip
void SurfaceImpl::DrawTextClipped(PRectangle rc, Scintilla::Font &font_, XYPOSITION ybase, const char *s, int len,
                                  ColourDesired fore, ColourDesired back) {
   FillRectangle(rc, back);
   DrawTextBase(rc, font_, ybase, s, len, fore);
}

void SurfaceImpl::DrawTextTransparent(PRectangle rc, Scintilla::Font &font_, XYPOSITION ybase, const char *s, int len,
                                  ColourDesired fore) {
   // Avoid drawing spaces in transparent mode
   for (int i=0;i<len;i++) {
      if (s[i] != ' ') {
         DrawTextBase(rc, font_, ybase, s, len, fore);
         return;
      }
   }
}

void SurfaceImpl::MeasureWidths(Scintilla::Font &font_, const char *s, int len, XYPOSITION *positions) {
   if (font_.GetID()) {
      Tk_Font tkfont = PFont(font_)->tkfont;
      XYPOSITION totalWidth = 0.0;
      for (int i = 0; i < len; i++) {
         int width = Tk_TextWidth(tkfont, (s+i), 1);
         totalWidth += width;
         positions[i] = totalWidth;
      }
   } else {
      // No font so return an ascending range of values
      for (int i = 0; i < len; i++) {
         positions[i] = i + 1.0;
      }
   }
}

XYPOSITION SurfaceImpl::WidthText(Scintilla::Font &font_, const char *s, int len) {
	if (font_.GetID()) {
            return (XYPOSITION)Tk_TextWidth(PFont(font_)->tkfont, s, len);
	} else {
		return 10.0;
	}
}

XYPOSITION SurfaceImpl::WidthChar(Scintilla::Font &font_, char ch) {
   return WidthText(font_, &ch, 1);
}

// Ascent and descent determined by Pango font metrics.

XYPOSITION SurfaceImpl::Ascent(Scintilla::Font &font_) {
   if (!(font_.GetID())) {
      return 1.0;
   }
   Tk_FontMetrics fm;
   Tk_GetFontMetrics(PFont(font_)->tkfont, &fm);
   return (XYPOSITION)fm.ascent;
}

XYPOSITION SurfaceImpl::Descent(Scintilla::Font &font_) {
	if (!(font_.GetID()))
		return 1.0;
   Tk_FontMetrics fm;
   Tk_GetFontMetrics(PFont(font_)->tkfont, &fm);
   return (XYPOSITION)fm.descent;
}

XYPOSITION SurfaceImpl::InternalLeading(Scintilla::Font &) {
   return 0.0;
}

XYPOSITION SurfaceImpl::ExternalLeading(Scintilla::Font &) {
   return 0.0;
}

XYPOSITION SurfaceImpl::Height(Scintilla::Font &font_) {
   return Ascent(font_) + Descent(font_);
}

XYPOSITION SurfaceImpl::AverageCharWidth(Scintilla::Font &font_) {
   return WidthChar(font_, 'n');
}

void SurfaceImpl::SetClip(PRectangle rc) {
//TODO: is this something that needs to be implemented??
}

void SurfaceImpl::FlushCachedState() {}

void SurfaceImpl::SetUnicodeMode(bool unicodeMode_) {
   if (unicodeMode_)
      et = UTF8;
}

void SurfaceImpl::SetDBCSMode(int codePage) {
   if (codePage && (codePage != SC_CP_UTF8))
      et = dbcs;
}

// The editor code calls this static function to get a surface object
Surface *Surface::Allocate(int) {
   Surface *surf = new SurfaceImpl();
//printf("Surface::Allocate -- THIS=%p\n",(void *)surf);
   return surf;
}


#ifdef SCI_LEXER

/* we can't get TclpFindSymbol from the TCL library because it's not exported. The following is a copy of the code
 */
#ifdef _WIN32
/*
 *----------------------------------------------------------------------
 *
 * TclpFindSymbol --
 *
 *  Looks up a symbol, by name, through a handle associated with a
 *  previously loaded piece of code (shared library).
 *
 * Results:
 *  Returns a pointer to the function associated with 'symbol' if it is
 *  found. Otherwise returns NULL and may leave an error message in the
 *  interp's result.
 *
 *----------------------------------------------------------------------
 */

static Tcl_PackageInitProc *
sTclpFindSymbol(
    Tcl_Interp *interp,
    Tcl_LoadHandle loadHandle,
    CONST char *symbol)
{
    Tcl_PackageInitProc *proc = NULL;
    HINSTANCE handle = (HINSTANCE)loadHandle;

    /*
     * For each symbol, check for both Symbol and _Symbol, since Borland
     * generates C symbols with a leading '_' by default.
     */

    proc = (Tcl_PackageInitProc *) GetProcAddress(handle, symbol);
    if (proc == NULL) {
    Tcl_DString ds;

    Tcl_DStringInit(&ds);
    Tcl_DStringAppend(&ds, "_", 1);
    symbol = Tcl_DStringAppend(&ds, symbol, -1);
    proc = (Tcl_PackageInitProc *) GetProcAddress(handle, symbol);
    Tcl_DStringFree(&ds);
    }
    return proc;
}

#endif

#ifdef  __linux

#include "dlfcn.h"

/*
 *----------------------------------------------------------------------
 *
 * TclpFindSymbol --
 *
 *	Looks up a symbol, by name, through a handle associated with a
 *	previously loaded piece of code (shared library).
 *
 * Results:
 *	Returns a pointer to the function associated with 'symbol' if it is
 *	found. Otherwise returns NULL and may leave an error message in the
 *	interp's result.
 *
 *----------------------------------------------------------------------
 */

static Tcl_PackageInitProc *
sTclpFindSymbol(
    Tcl_Interp *interp,		/* Place to put error messages. */
    Tcl_LoadHandle loadHandle,	/* Value from TcpDlopen(). */
    CONST char *symbol)		/* Symbol to look up. */
{
    CONST char *native;
    Tcl_DString newName, ds;
    VOID *handle = (VOID*)loadHandle;
    Tcl_PackageInitProc *proc;

    /*
     * Some platforms still add an underscore to the beginning of symbol
     * names. If we can't find a name without an underscore, try again with
     * the underscore.
     */

    native = Tcl_UtfToExternalDString(NULL, symbol, -1, &ds);
    proc = (Tcl_PackageInitProc *) dlsym(handle,	/* INTL: Native. */
	    native);
    if (proc == NULL) {
	Tcl_DStringInit(&newName);
	Tcl_DStringAppend(&newName, "_", 1);
	native = Tcl_DStringAppend(&newName, native, -1);
	proc = (Tcl_PackageInitProc *) dlsym(handle,	/* INTL: Native. */
		native);
	Tcl_DStringFree(&newName);
    }
    Tcl_DStringFree(&ds);

    return proc;
}

#endif


class DynamicLibraryImpl : public DynamicLibrary {
protected:
    Tcl_LoadHandle handle;
    Tcl_FSUnloadFileProc *unloadProc;
public:
    DynamicLibraryImpl(const char *modulePath) {
        // convert modulePath to Tcl_Obj
        Tcl_Obj *pathPtr = Tcl_NewStringObj(modulePath, strlen(modulePath));
        Tcl_PackageInitProc *unused;
        Tcl_IncrRefCount(pathPtr);
        if (Tcl_FSLoadFile(scintilla_interp, pathPtr, "unused1", "unused2", &unused, &unused, &handle, &unloadProc) != TCL_OK) {
            fprintf(stderr,"File load of %s failed %s\n", modulePath, Tcl_GetStringResult(scintilla_interp));
            handle = NULL;
        }
        Tcl_DecrRefCount(pathPtr);
    }

    virtual ~DynamicLibraryImpl() {
    }

    // Use g_module_symbol to get a pointer to the relevant function.
    virtual Function FindFunction(const char *name) {
        if (handle != NULL) {
            void *ptr = (void*) sTclpFindSymbol(scintilla_interp, handle, name);
            if (ptr) {
                return ptr;
            } else
                return NULL;
        } else
            return NULL;
    }

    virtual bool IsValid() {
        return handle != NULL;
    }
};

DynamicLibrary *DynamicLibrary::Load(const char *modulePath) {
   return static_cast<DynamicLibrary *>( new DynamicLibraryImpl(modulePath) );
}
#endif


ColourDesired Platform::Chrome() {
   return ColourDesired(0xe0, 0xe0, 0xe0);
}

ColourDesired Platform::ChromeHighlight() {
   return ColourDesired(0x00, 0x00, 0xff);
//   return ColourDesired(0xff, 0xff, 0xff);
}

const char *Platform::DefaultFont() {
   return "courier";
}

int Platform::DefaultFontSize() {
   return 12;
}

unsigned int Platform::DoubleClickTime() {
   return 500;    // Half a second
}

bool Platform::MouseButtonBounce() {
   return true;
}

void Platform::DebugDisplay(const char *s) {
   fprintf(stderr, "%s", s);
}

bool Platform::IsKeyDown(int) {
   // TODO: discover state of keys in GTK+/X
   return false;
}

long Platform::SendScintilla(
    WindowID w, unsigned int msg, unsigned long wParam, long lParam) {
   return 0;//scintilla_send_message(SCINTILLA(w), msg, wParam, lParam);
}

long Platform::SendScintillaPointer(
    WindowID w, unsigned int msg, unsigned long wParam, void *lParam) {
   return 0;
//    return scintilla_send_message(SCINTILLA(w), msg, wParam,
//                                  reinterpret_cast<sptr_t>(lParam));
}

bool Platform::IsDBCSLeadByte(int codePage, char ch) {
   // Byte ranges found in Wikipedia articles with relevant search strings in each case
   unsigned char uch = static_cast<unsigned char>(ch);
   switch (codePage) {
      case 932:
         // Shift_jis
         return ((uch >= 0x81) && (uch <= 0x9F)) ||
            ((uch >= 0xE0) && (uch <= 0xFC));
            // Lead bytes F0 to FC may be a Microsoft addition. 
      case 936:
         // GBK
         return (uch >= 0x81) && (uch <= 0xFE);
      case 950:
         // Big5
         return (uch >= 0x81) && (uch <= 0xFE);
      // Korean EUC-KR may be code page 949.
   }
   return false;
}

int Platform::DBCSCharLength(int codePage, const char *s) {
   if (codePage == 932 || codePage == 936 || codePage == 950) {
      return IsDBCSLeadByte(codePage, s[0]) ? 2 : 1;
   } else {
      int bytes = mblen(s, MB_CUR_MAX);
      if (bytes >= 1)
         return bytes;
      else
         return 1;
   }
}

int Platform::DBCSCharMaxLength() {
   return MB_CUR_MAX;
   //return 2;
}

// These are utility functions not really tied to a platform

int Platform::Minimum(int a, int b) {
   if (a < b)
      return a;
   else
      return b;
}

int Platform::Maximum(int a, int b) {
   if (a > b)
      return a;
   else
      return b;
}

//#define TRACE

#ifdef TRACE
void Platform::DebugPrintf(const char *format, ...) {
   char buffer[2000];
   va_list pArguments;
   va_start(pArguments, format);
   vsprintf(buffer, format, pArguments);
   va_end(pArguments);
   Platform::DebugDisplay(buffer);
}
#else
void Platform::DebugPrintf(const char *, ...) {}

#endif

// Not supported for GTK+
static bool assertionPopUps = true;

bool Platform::ShowAssertionPopUps(bool assertionPopUps_) {
   bool ret = assertionPopUps;
   assertionPopUps = assertionPopUps_;
   return ret;
}

void Platform::Assert(const char *c, const char *file, int line) {
   char buffer[2000];
   sprintf(buffer, "Assertion [%s] failed at %s %d", c, file, line);
   strcat(buffer, "\r\n");
   Platform::DebugDisplay(buffer);
   abort();
}

int Platform::Clamp(int val, int minVal, int maxVal) {
   if (val > maxVal)
      val = maxVal;
   if (val < minVal)
      val = minVal;
   return val;
}

void Platform_Initialise() {
   FontMutexAllocate();
}

void Platform_Finalise() {
   FontMutexFree();
}


#ifdef SCI_NAMESPACE
namespace Scintilla {
#endif
//==========================================================================
// CLASS Menu implementations (that seem to be needed from based class
// to get symbols to resolve during library loading)
//
Menu::Menu() : mid(0) {}
void Menu::Destroy() {
}
void Menu::CreatePopUp() {
}
void Menu::Show(Point pt, Window &) {
}

//==========================================================================
// CLASS Font implementation (also seemingly required)
//
Font::Font() : fid(0) {}
Font::~Font() {}

static Tcl_HashTable *fontTable = NULL;

void Font::Create(const FontParameters &fp) {
Tcl_HashEntry *hPtr = NULL;
int isNew;

//fprintf(stderr,"Font::Create ((%p))--> %s / %f / %d / italic=%d\n",(void *)this,fp.faceName,fp.size,fp.weight,fp.italic);

    if (fontTable == NULL) {
        // cache fonts to improve performance
        fontTable = new Tcl_HashTable;
        Tcl_InitHashTable(fontTable, TCL_STRING_KEYS);
    }

//TODO: NEED THE tkwin REFERENCE FOR THIS!!
//TODO: FIGURE OUT HOW TO SPECIFY THE Tk_Font BASED FROM FontParameters
//NOTE: The FontParameters gets family/size from Platform::DefaultFont
//      and Platform::DefaultFontSize

   // check if the point size is encoded to represent a negative value, which
   // for Tk would mean it will use that value as the number of pixels.
   int ptsize = static_cast<int>(fp.size);
   if (ptsize > MAGIC_FONT_OFFSET)
       ptsize = (MAGIC_FONT_OFFSET - ptsize);

   //########################################################
   // Font format: <family name> ?pt size? ?bold? ?italic?
   //########################################################
#if (TK_MAJOR_VERSION == 8 && TK_MINOR_VERSION < 5)
   Tcl_Obj *objPtr = Tcl_NewObj();
   char dbuf[64];
   sprintf(dbuf," %d", ptsize);
   Tcl_AppendStringsToObj(objPtr,"{", fp.faceName, "}", dbuf, (char*)NULL);
#else
   Tcl_Obj *objPtr = Tcl_ObjPrintf("{%s} %d", fp.faceName, ptsize);
#endif

   if (fp.weight == SC_WEIGHT_BOLD) {
      Tcl_AppendStringsToObj(objPtr, " bold", (char *)NULL);
   }
   if (fp.italic) {
      Tcl_AppendStringsToObj(objPtr, " italic", (char *)NULL);
   }

   // Use hash table of previously created font descriptions as Tcl_Obj*
   // so we can realize a performance boost in Tk_AllocFontFromObj which
   // caches information in the Tcl_Obj.
   hPtr = Tcl_CreateHashEntry(fontTable, Tcl_GetString(objPtr), &isNew);
   FontHandle *p;
   if (isNew) {
      Tk_Font tkf = Tk_GetFont(Tk_Interp(_curr_tkwin),_curr_tkwin, Tcl_GetString(objPtr));
      p = new FontHandle(tkf);
      Tcl_SetHashValue(hPtr, new FontHandleMgr(p));
   } else {
      FontHandleMgr *mgrp = (FontHandleMgr *)Tcl_GetHashValue(hPtr);   // reuse existing object
      p = mgrp->GetHandle();
   }

   fid = new FontHandleMgr(p);
}

void Font::Release() {
   if (fid) {
      delete reinterpret_cast<FontHandleMgr *>(fid);
   }
   fid = 0;
}

//==========================================================================
// CLASS Window implementation (also seemingly required)
//
Window::~Window() {}
void Window::Destroy() {}

PRectangle Window::GetMonitorRect(Point pt) {
   return PRectangle(0,0,0,0);
}

void Window::SetCursor(Cursor curs) {

   ScintillaObject *sc_obj = reinterpret_cast<ScintillaObject *>(wid);
   if (!sc_obj) return;

   if ( ! Tk_IsMapped(sc_obj->tkwin) ) return;

//printf("##################Window::SetCursor %d\n",curs);
   TkSciExt::MSG_SetCursor((ClientData)sc_obj->client, static_cast<int>(curs));
}
void Window::SetFont(Scintilla::Font &) {
}
void Window::Show(bool show) {
}
void Window::SetPosition(PRectangle rc) {
}
void Window::SetPositionRelative(PRectangle rc, Window relativeTo) {
}
PRectangle Window::GetPosition() {
   return PRectangle(0,0,0,0);
}
PRectangle Window::GetClientPosition() {
   return GetPosition();
}
void Window::InvalidateRectangle(PRectangle rc) {

    ScintillaObject *sc_obj = reinterpret_cast<ScintillaObject *>(wid);
    if (!sc_obj) {
       return;
    }
    if ( ! Tk_IsMapped(sc_obj->tkwin) ) {
//printf("Window::InvalidateRectangle NOT MAPPED YET for sc_obj=%p\n",(void *)sc_obj);
       return;
    }
//printf("Window::InvalidateRectangle((%p)) >> +%d+%dx%dx%d\n",(void *)this,rc.left,rc.top,(rc.right-rc.left),(rc.bottom-rc.top));
    // We use an Expose event to indicate to the Tk widget functionality
    // what region needs to be repainted
    XEvent e;
    e.xexpose.type    = Expose;
    e.xexpose.count   = -1;
    e.xexpose.display = Tk_Display(sc_obj->tkwin);
    e.xexpose.window  = Tk_WindowId(sc_obj->tkwin);
    // HACK ALERT(maybe) We manually force the expose region to include
    // any margins which are visible
    if (rc.left > 0) {
       rc.left = 0;
       rc.right++;
    }
    e.xexpose.x       = static_cast<int>(rc.left);
    e.xexpose.y       = static_cast<int>(rc.top);
    e.xexpose.width   = static_cast<int>(rc.right - rc.left);
    e.xexpose.height  = static_cast<int>(rc.bottom - rc.top);

    Tk_QueueWindowEvent(&e, TCL_QUEUE_TAIL);
}
void Window::InvalidateAll() {
printf("Window::InvalidateAll\n");
}

//==========================================================================
// CLASS ListBox implementation (also seemingly required)
// NOTE: even if calltips and code-completion aren't going to be used, the
// following class _IS_ required
//
class ListBoxX : public ListBox {
public:
   ListBoxX() {}
   virtual ~ListBoxX() {}
   void SetFont(Font &) {}
   void Create(Window &, int, Point, int, bool, int) {}
   void SetAverageCharWidth(int) {}
   void SetVisibleRows(int) {}
   int GetVisibleRows() const {return 0;}
   PRectangle GetDesiredRect() { return PRectangle(); }
   int CaretFromEdge() { return 0; }
   void Clear() {}
   void Append(char *s, int type = -1) {}
   int Length() { return 0; }
   void Select(int n) {}
   int GetSelection() { return 0; }
   int Find(const char *prefix) { return -1; }
   void GetValue(int n, char *value, int len) {}
   void RegisterImage(int type, const char *xpm_data) {}
   void RegisterRGBAImage(int, int, int, const unsigned char *) {}
   void ClearRegisteredImages() {}
   void SetDoubleClickAction(CallBackAction, void *) {}
   void SetList(const char* list, char separator, char typesep) {}
};
ListBox::ListBox() {}
ListBox::~ListBox() {}
ListBox *ListBox::Allocate() {
   return new ListBoxX();
}

//==========================================================================
// CLASS ElapsedTime implementation (also seemingly required)
ElapsedTime::ElapsedTime() {
}
#ifdef SCI_NAMESPACE
}
#endif
