//
// scintilla-cmd.h --
//
// Declarations shared among the files that implement the Tk-flavor
// of the Scintilla editor widget.
//
// Copyright (c) 2013 Mentor Graphics Corporation
// The license.terms file describes the conditions under which this software may be distributed.

#ifndef _SCINTILLA_CMD
#define _SCINTILLA_CMD 

#include <string.h>

// NOTE: Because the Scintilla code is C++, we need for this Tcl widget
//       to use a C++ compiler
#include "Scintilla.h"
#include <SciLexer.h>
#include "ScintillaTk.h"
#include <bitset>
#include <vector>

#include "Platform.h" // for access to PRectangle
using namespace Scintilla;

// create a namespace for ourselves
namespace TkSci {

//----------------------------------------------------------------
// We do _NOT_ want the following within the TkSci namespace.

// next one needs C access because it's called as an idle task
extern "C" void SciWidgetDisplay (ClientData clientData);

#ifdef _WIN32
extern "C" __declspec(dllexport) int Scintillatk_Init(Tcl_Interp *interp);
#endif
//----------------------------------------------------------------



// define mapping macros to go between Tk's 1-based line number
// and Scintilla's 0-based line numbers
#define LINE2TK(ln)  ((ln < 0) ? ln : (ln + 1))
#define LINE2SCI(ln) ((ln < 0) ? ln : (ln - 1))

// Flag bit definitions.
//
#define REDRAW_PENDING 0x1
#define GOT_FOCUS      0x2
#define REDRAW_YSCROLL 0x4
#define REDRAW_XSCROLL 0x8
#define REDRAW_EXPOSE  0x10
#define INIT_CFG_DONE  0x20
#define BUSY_MODE      0x40
#define BUTTON_DOWN    0x80
#define DESTROYED      0x100
#define RESIZE         0x120
#define GOT_SELECTION  0x200

// data structure describing a change in the
// number of lines within the widget
typedef struct {
   int line;  // the line location of the add/delete
   int added; // >0 to indicate an addition, <0 indicates a deletion
} SciWidgetLinesAdded;

// forward decl
class MarginMgr;
class TagMgr;
class MarkerMgr;
class MarkMgr;


// data structure for widget
typedef struct {
   ScintillaObject *editor;
   Tcl_Interp      *interp;
   Tcl_Command      widgetCmd;
   Tk_Window        tkwin;
   Display         *display;
   GC               copyGC;
   Pixmap           mainPixmap;
   int				width;
   int              height;
   Tk_OptionTable   optionTable;
   Tk_BindingTable  bindingTable;

   /* widget-specific options */
   Tk_Font     tkfont;
   Tk_Cursor   cursor;
   Tk_Cursor   busycursor;
   Tk_3DBorder background;
   XColor     *foreground;
   XColor     *highlight;
   XColor     *highlightBg;
   XColor     *marginBg;
   XColor     *marginFg;
   int         relief;
   int         highlightWidth;
   int         borderWidth;
	int        reqWidth;
	int        reqHeight;
   char       *showws;
   char       *takefocus;
   int         tabwidth;
   int         state;
   bool        readonly;
   char       *language;
   int         langType;
   int         foldstyle;
   char        undo;
   char        autoseps; // only provided to be compatible w/ Tk Text widget
   char        inUndoAction;    // keeps track of undo collection state when "edit separator" used
   char       *blinkcaret;
   bool        exportSelection; // controls if selection is X global selection

   /* callback commands */
   char       *linesAddedCmd;
   char       *marginWidthCmd;
   char       *moveUpdateCmd;
   char       *yScrollCmd;
   char       *xScrollCmd;

   /* other stuff */
   int        padX;
   int        padY;
   int        yscrollDelta;
   int        flags;
   MarginMgr *marginMgr;
   MarkerMgr *markerMgr;
   MarkMgr   *markMgr;
   TagMgr    *tagMgr;
   SciWidgetLinesAdded *added_data;
   int       styleBits; // number of bits in styleId devoted to style, remainder of 8 bits are indicator bits
   Tcl_Obj   *keywords[KEYWORDSET_MAX];

   /* data used for expose repaints */
   int expose_x;
   int expose_y;
   int expose_w;
   int expose_h;
} SciWidget;


// provide a callback mechanism for code on the Scintilla side
#define SCIW_SYNC_XSCROLL      100
#define SCIW_SYNC_YSCROLL      101
#define SCIW_MOVE_UPDATE       102
#define SCIW_LINESADDED_UPDATE 103
#define SCIW_MODIFIED          104

extern void MessageFromScintilla(ClientData clientData, int msg_id, uptr_t data);
extern void MessageFromScintilla(ClientData clientData, int msg_id);


////////////////////////////////////////////////////////////////////
// Margin management
//
enum MarginType {
   NONE, LINE_NUMS, LINE_NUMS_ALT, TEXT, RTEXT, SYMBOL, FOLD
};
static const char *MarginTypeNames[] = {
   "none", "lnums", "lnums_alt", "text", "rtext", "symbol", "fold"
};
#define IS_LINE_NUMS(t) (t == LINE_NUMS || t == LINE_NUMS_ALT)
#define IS_TEXT_TYPE(t) (IS_LINE_NUMS(t) || t == TEXT || t == RTEXT)

// data structure to hold configuration options used by each margin
typedef struct {
   char     *alias;     //(STRING) an alternate name(ie alias) for identifying the margin
   char     *clickcmd;  //(STRING) command to callback for clicks (null = clicks disabled)
   int       type;      //(STRING_TABLE) index of margin type in MarginTypeNames
   int       width;     //(INT)    margin's width(in pixels) - value of "0" will hide the margin
   Tk_Cursor cursor;    //(CURSOR) margin's cursor
} MarginOptions;

//-------------------------------------
// CLASS MarginData
//-------------------------------------
class MarginData {
public:
   MarginData(MarginMgr *, int, int);
   ~MarginData();

   // these can't be changed by client going thru the 
   // options interface, so we keep them as class data
   //
   int col_id;   // Scintilla uses columns 0 - 4
   int viswidth; // holds width when margin is visible

   // other info
   MarginMgr     *mgr;
   MarginOptions *mopts;
   Tk_OptionTable optionTable;

   // functions
   int Configure(int, Tcl_Obj *CONST[]);

   // accessor calls for options data
   const char *GetAlias()    { return mopts->alias; }
   const char *GetClickCmd() { return mopts->clickcmd; }
   const char *GetTypeStr()  { return MarginTypeNames[mopts->type]; }
   MarginType  GetType()     { return static_cast<MarginType>(mopts->type); }
   Tk_Cursor   GetCursor()   { return mopts->cursor; }
   const char *GetCursorStr();
   int  GetWidth()           { return mopts->width; }
   int  GetVisibleWidth()    { return viswidth; }
   void SetWidth(int w)      { mopts->width = w; viswidth = w;}
};


//-------------------------------------
// CLASS MarginMgr
//-------------------------------------
class MarginMgr {
public:
   MarginMgr(SciWidget *);
   ~MarginMgr();

   // these support the command interface to the widget
   int  CmdBbox    (int objc, Tcl_Obj *CONST objv[]);
   int  CmdCget    (int objc, Tcl_Obj *CONST objv[]);
   int  CmdConfig  (int objc, Tcl_Obj *CONST objv[]);
   int  CmdDump    (int objc, Tcl_Obj *CONST objv[]);
   int  CmdFillNums(int objc, Tcl_Obj *CONST objv[]);
   int  CmdIndex   (int objc, Tcl_Obj *CONST objv[]);
   int  CmdMarker  (int objc, Tcl_Obj *CONST objv[]);
   int  CmdShow    (int objc, Tcl_Obj *CONST objv[], bool);
   int  CmdTag     (int objc, Tcl_Obj *CONST objv[]);
   int  CmdText    (int objc, Tcl_Obj *CONST objv[]);
   int  CmdVisible (int objc, Tcl_Obj *CONST objv[]);
   int  InstanceCmd(int objc, Tcl_Obj *CONST objv[]);

   void Click(XButtonEvent be);
   void DumpText(Tcl_DString *dstr, int line1, int idx1, int line2, int idx2, const char* command);
   int  Get(Tcl_Obj *, int *);
   int  GetLineNWidth(int ndigits = 0);
   void RemoveMarker(int id);
   void RemoveMarker(int marginID, int id);
   void RemoveAllMarkers(int marginID);
   bool SetCursor(int x, int y);
   void UpdateMargin(int);
   int  ValidateTypeChange(int);
   int  Width();
   void ZoomPre();
   void ZoomPost();

private:
   //--- VARIABLES
   static const int MAXM = 5; // largest margin id

   MarginData    *margins[MAXM];
   SciWidget     *sciwidgetPtr; // note: required var name for use by SSM macro

public:
   // give easy access to MarginData
   Tcl_Interp    *interp;
   Display       *display;
   Tk_Window      tkwin;

private:
   // linenumber and fold columns are special. make it easy to find them
   int currLnum;
   int currFold;

   int preZoomWidth; // holds char width prior to a zoom

   //--- FUNCTIONS
   int  _atX(int x);
   void _click(int, int, int, int, int, bool, bool, bool);
   bool _hasMarker(int, int);
   int  _get(Tcl_Obj *, int *);
   void _reportError(const char *);
   void _reportError(Tcl_Obj *);
   int  _show(int id, bool show);
   int  _validAlias(const char *);
   int  _validId(int);
   int  _validName(const char *, int *);
   int  _validType(const char *, MarginType *);
};


////////////////////////////////////////////////////////////////////
// Tag management

typedef enum {
    TT_TEXT  = 1,
    TT_MARG  = 2,
    TT_ANNO  = 4,
    TT_INDIC = 8
} TagType;

// tags can be associated with an indicator style (they match those
// defined by Scintilla, but aren't hardwired to their constant values)
typedef enum {
    indicNONE,
    indicPlain,         // solid 1-pixel underline
    indicSquiggle,      // squiggly 3-pixel high underline
    indicTt,            // small "T" shapes drawn under text
    indicDiag,          // diaganol hatching
    indicStrike,        // strike-through drawn at midpoint of text
    indicHidden,        // an indicator that has no visual attributes
    indicBox,           // rectangle drawn around text using solid 1-pixel line
    indicRoundBox,      // rectange w/ rounded corners drawn around text - uses transparent interior
    indicSquareBox,     // same as above but with square corners
    indicDash,          // dashed underline
    indicDots,          // dotted underline
    indicSquiggleLow,   // same as indicSquiggle but only using 2 pixels
    indicDotBox,        // dotted rectangle around text - uses transparent interior
    indicSquigglePixMap // dotted rectangle around text - uses transparent interior
} IndicTypeID;

// Structure defining a tag location
typedef struct {
    sptr_t startpos;   // position within document where tag begins
    sptr_t endpos;     // position within document where tag ends
} Location;

class TagLocation;
class TagMgr;

// options for TagData, in struct so Tk_Offset can be used to set options
typedef struct {
    XColor        *fgColor;    // foreground color of text
    XColor        *bgColor;    // background color of text
    int            bold;       // boolean to indicate if text is bold
    int            eolfilled;  // boolean to indicate if background color of last character in line extends to right margin
    int            italic   ;  // boolean to indicate if text is italicized
    int            underline;  // boolean to indicate if text is underlined
    IndicTypeID    indicType;  // indicator type (if any)
    int            alpha;      // alpha value to be used for transparent indicators
    int            alphab;     // alpha value to be used for the boarder of transparent indicators
} TagDataOptions;

// data structure to hold configuration options used by each tag
class TagData {
    typedef enum {
        FLAG_NONE = 0,
        FLAG_DELETING = 1
    } TagDataFlags;
public:
    char    *name;             // symbolic name of the tag
    int            tagType;    // mask intended to identify the owning area (text or margin)
    int            style_id;   // Scintilla style id
    int            indic_id;   // Scintilla id used for indicators
    TagDataOptions options;
    Tk_OptionTable optionTable;// table to hold config options

    std::vector<TagLocation*> tagLocations; // sorted array of locations where tag applied, we don't "own" TagLocation* pointers, don't delete them!
    int            priority;   // priority relative to other tags when more than one tag applied to same text
    int            flags;      // flags

    TagData(Tcl_Interp *interp, const char* n, int ttMask, int _priority);
    ~TagData();

    const char *Name()      { return name; }
    int Bold()              { return options.bold; }
    int EolFilled()         { return options.eolfilled; }
    int Italic()            { return options.italic; }
    int Underline()         { return options.underline; }
    XColor *BG()            { return options.bgColor; }
    XColor *FG()            { return options.fgColor; }
    int StyleID()           { return style_id; }
    int IndicID()           { return indic_id; }
    int Type()              { return tagType; }
    IndicTypeID IndicType() { return options.indicType; }
    int Alpha()             { return options.alpha; }
    int AlphaB()            { return options.alphab; }
    void RegisterLocation(TagLocation* t_p);
    void UnregisterLocation(TagLocation* t_p);
    void NextRange(Tcl_Interp *interp, ScintillaObject *obj, sptr_t pos1, sptr_t pos2, Tcl_Obj* indices);
    void PrevRange(Tcl_Interp *interp, ScintillaObject *obj, sptr_t pos1, sptr_t pos2, Tcl_Obj* indices);
    void Ranges(Tcl_Interp *interp, ScintillaObject *obj, Tcl_Obj* indices);
    void RemoveTag(TagMgr *mgr, sptr_t startpos, sptr_t endpos);
    void DumpLocations();
    void DeleteTag(TagMgr *mgr);
};

class TagLocation {
    Location location;
    std::vector<TagData*> tags; // list of tags at location, we don't "own" TagData* pointers, don't delete them!
public:
    TagLocation(TagData *tag, sptr_t start, sptr_t end);
    TagLocation(TagLocation *, sptr_t end);
    ~TagLocation();
    // associate tag with location, return 1 if tag's priority is higher than previous high and display needs update
    int RegisterTag(TagData *td);
    // disassociate tag with location, return new highest priority tag if display needs update (td was highest priority)
    TagData *UnregisterTag(TagData *td);
    void setStartPos(sptr_t start);
    void setEndPos(sptr_t end);
    sptr_t startpos()   { return location.startpos; }
    sptr_t endpos()   { return location.endpos; }
    bool empty()   { return tags.empty(); }
    void dumpTags();
};

//-------------------------------------
// CLASS TagMgr
//-------------------------------------
class TagMgr {
    friend class TagData;
public:
   TagMgr(SciWidget *);
   ~TagMgr();

   // these support the command interface to the widget
   int  CmdAdd     (int objc, Tcl_Obj *CONST objv[]);
   int  CmdBind    (int objc, Tcl_Obj *CONST objv[]);
   int  CmdCget    (int objc, Tcl_Obj *CONST objv[]);
   int  CmdConfig  (int objc, Tcl_Obj *CONST objv[]);
   int  CmdDelete  (int objc, Tcl_Obj *CONST objv[]);
   int  CmdNames   (int objc, Tcl_Obj *CONST objv[]);
   int  CmdNPRange (int objc, Tcl_Obj *CONST objv[], bool next);
   int  CmdRanges  (int objc, Tcl_Obj *CONST objv[]);
   int  CmdRemove  (int objc, Tcl_Obj *CONST objv[]);
   int  InstanceCmd(int objc, Tcl_Obj *CONST objv[]);

   void      HandleBindEvent(XEvent *eventPtr);
   void      BindEvent(XEvent *eventPtr, int nTags, TagData **tagArrayPtr);
   void      DumpText(int mask, Tcl_DString *dstr, int pos1, int pos2, const char *command);
   void      DumpMargin(int mask, Tcl_DString *dstr, int line1, int pos1, int line2, int pos2, const char *command);
   TagData  *FindTag(const char *tagname);
   TagData  *FindTag(int style_id);
   void      FireSelectionVirtual();
   TagData  *GetMarginTag() { return marginTag; }
   TagData  *GetSelectTag() { return selTag; }
   void      PickCurrent(XEvent *);
   int       Reset(Tcl_Interp *interp, int langType);
   void      UpdateMarginColor(bool bg, bool fg);
   TagData  *ValidateForMargin(const char *tagname);
   bool IsIndicator(TagData *tagPtr) {
       return (tagPtr->Type() & TT_INDIC);
   }
   // return list of all margin tags
   Tcl_Obj *GetMarginTags(Tcl_Interp *interp);
   int       GetMarginStyleOffset() { return marginStyleOffset; }
   int       GetAnnotationStyleOffset() { return annotationStyleOffset; }

private:
   int       _configIndicator(TagData *);
   TagData  *_createTag(int ttMask, const char *tagname, int *newTag = NULL);
   void      _deleteTag(TagData *tagPtr, bool force = false);
   TagData **_getTagsAtPos(int pos, int *num);
   bool      _isTagAtPos(TagData *tagPtr, int pos, const char *tagName = NULL);
   int       _nextStyleID(int *, int ttMask);
   int       _nextIndicID(int *);
   void      _setIndicStyle(TagData *, int);

   // for testing purposes
   void _dump  (TagData *tagPtr = NULL);
   void _dumpIt(TagData *tagPtr);

   const char *TAG_SEL;
   bool IS_SEL(const char *tagname) {
      return (strcmp(tagname, TAG_SEL) == 0);
   }
   const char *TAG_MARGIN;
   bool IS_MARGIN(const char *tagname) {
      return (strcmp(tagname, TAG_MARGIN) == 0);
   }

   void resetFreeTextStyles();
   void resetFreeMarginStyles();
   void resetFreeAnnotationStyles();
   void resetFreeIndicators();
   void deleteTag(const char* name);

   SciWidget     *sciwidgetPtr;
   Tcl_Interp    *interp;
   Tk_Window      tkwin;

   Tcl_HashTable tagTable;
   TagData      *selTag;      // special built-in tag for selection
   TagData      *marginTag;   // special built-in tag for the margins
   int           freeIndicators; // bit mask of available indicators (max 32)
   std::bitset<256>  freeTextStyles;
   std::bitset<256>  freeMarginStyles;
   std::bitset<256>  freeAnnotationStyles;
   int           numTags;     // total number of defined tags
   int           annotationStyleOffset;
   int           marginStyleOffset;

   // info used for event bindings associated with tags
   XEvent pickEvent;           /* The event used to determine "current" character */
   int numCurTags;             /* Number of tags associated with character at
                                * current mark. */
   TagData **curTagArrayPtr;   /* Pointer to array of tags for current mark,
                                * or NULL if none. */

    // tag location management for fast lookup of tag by position
    std::vector<TagLocation*> tagLocations; // sorted array of tag locations

    void ApplyStyle(TagData *td, sptr_t startpos, sptr_t endpos);
    void RemoveStyle(TagData *td, sptr_t startpos, sptr_t endpos);
    void ApplyTag(TagData *td, sptr_t startpos, sptr_t endpos);
    void RemoveTag(TagData *td, sptr_t startpos, sptr_t endpos);
    void RemoveLocation(TagLocation *tl_p);
};


////////////////////////////////////////////////////////////////////
// Marker management

// data structure for marker configuration options
typedef struct {
   XColor *bgColor; //(COLOR) bgd color of marker
   XColor *fgColor; //(COLOR) fgd color
   char   *imgName; //(STRING) name of Tk image - only used w/ -image option
   Tk_Image img;    //(Tk_Image) image handle
   int     type;    //(STRING_TABLE) index of marker type in MarkerTypeNames
    const unsigned char *pixels;
} MarkerOptions;

//-------------------------------------
// CLASS MarkerData
//-------------------------------------
class MarkerData {
public:
   MarkerData(MarkerMgr *);
   ~MarkerData();

   // these can't be changed by client going thru the 
   // options interface, so we keep them as class data
   //
   char *name;    // name to identify the marker
   int id;        // logical ID identifying bit-mask position
   int handle;    // Scintilla handle created when this marker is assigned
   int typeID;    // Scintilla constant (eg SC_MARK_xxx)
   int margin;    // ID of margin holding this marker
   int cloned_n;  // the count of how many cloned copies there are

   MarkerData *cloned_from; // is cloned, this will be the marker it was cloned from

   // other info
   MarkerMgr     *mgr;
   MarkerOptions *mopts;
   Tk_OptionTable optionTable;

   void UpdateType();
   void Copy(MarkerData *src);
   int CloneCount() { return cloned_n; }
   bool IsClone() { return (cloned_from != NULL); }
   bool MatchName(const char *src_name);
};

//-------------------------------------
// CLASS MarkerMgr
//-------------------------------------
typedef enum {
    cloneCopy,  // 0
    cloneDelete // 1
} CloneOpId;

class MarkerMgr {
public:
   MarkerMgr(SciWidget *);
   ~MarkerMgr();

   // these support the command interface to the widget
   int  CmdConfig  (int objc, Tcl_Obj *CONST objv[]);
   int  CmdDelete  (int objc, Tcl_Obj *CONST objv[]);
   int  CmdExists  (int objc, Tcl_Obj *CONST objv[]);
   int  CmdNames   (int objc, Tcl_Obj *CONST objv[], bool include_clones = false);
   int  InstanceCmd(int objc, Tcl_Obj *CONST objv[]);

   int         AddMarker(int marginID, MarkerData *mPtr, int line);
   MarkerData *FindMarker(const char *name);
   MarkerData *FindMarker(int marker_id);
   MarkerData *FindMarker(int marginID, const char *, bool ok_to_clone = true);
   int         GetMaxId();

   int         _validateImage(MarkerData *);

private:
   MarkerData *_cloneOp(CloneOpId, MarkerData *);
   MarkerData *_cloneMarker(const char *, MarkerData *);
   MarkerData *_createMarker(const char *, int *);
   void        _deleteMarker(MarkerData *);
   void        _deleteMarkerImpl(MarkerData *);
   int         _getNextId();

   // for testing purposes
   void _dump  (MarkerData *mPtr = NULL);
   void _dumpIt(MarkerData *mPtr);

   Tcl_HashTable  markerTable;
   int            ids;

public:
   // for easy access by MarkerData
   SciWidget     *sciwidgetPtr;
   Tcl_Interp    *interp;
   Tk_Window      tkwin;
};


////////////////////////////////////////////////////////////////////
// Mark management (this is the same as Tk Text widget "marks")

typedef struct {
    const char *name;
    int         pos;
    bool        gravityLeft; // TRUE = left, FALSE = right
} MarkData;

//-------------------------------------
// CLASS MarkMgr
//-------------------------------------
class MarkMgr {
public:
    MarkMgr(SciWidget *);
    ~MarkMgr();

    // these support the command interface to the widget
    int  CmdGravity (int objc, Tcl_Obj *CONST objv[]); 
    int  CmdSet     (int objc, Tcl_Obj *CONST objv[]); 
    int  InstanceCmd(int objc, Tcl_Obj *CONST objv[]);

    MarkData *AddMark(const char *name, int pos);
    MarkData *FindMark(const char *, int *offset = NULL);
    int       GetCurrentPos() { return currentMark->pos; }
    // dump marks in selected margin from line1 to line2
    void      Dump(Tcl_DString *dstr, int margin_id, int line1, int line2, const char *command);

private:
    MarkData *_createMark(const char *name, int *newMark = NULL);
    void      _deleteMark(MarkData *);

    // for testing purposes
    void _dump(MarkData *mPtr = NULL);

    SciWidget     *sciwidgetPtr;
    MarkData      *insertMark;  // special built-in mark
    MarkData      *currentMark; // special built-in mark
    Tcl_HashTable  markTable;
    Tcl_Interp    *interp;
    Tk_Window      tkwin;
};

} // end of "using namespace TkSci"
#endif
